﻿using System;
using System.ComponentModel;
using System.Net;
using System.Net.Sockets;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    public class SocketTcpClient
    {
        public event Action<RemoteInfo> OnConnect;
        public event Action<RemoteInfo> OnReceive;
        public event Action<RemoteInfo> OnDisconnect;
        public event Action<RemoteInfo> OnSend;

        private Socket socket;

        public Task ConnectAsync(string ipAddress, int port)
        {
            if (!IPAddress.TryParse(ipAddress, out IPAddress ip))
            {
                throw new Exception("Invalid ipAddress given");
            }

            return CreateSocketTcpClientAsync(ip, port);
        }

        private async Task CreateSocketTcpClientAsync(IPAddress ipAddress, int port)
        {
            IsIP6 = ipAddress.AddressFamily == AddressFamily.InterNetworkV6;
            socket = new Socket(IsIP6 ? AddressFamily.InterNetworkV6 : AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            try
            {
                await socket.ConnectAsync(ipAddress, port).ConfigureAwait(false);
                OnConnect?.Invoke(new RemoteInfo { IsTcp = true, IsIpv6 = IsIP6 });
                _ = StartReceivingAsync();
            }
            catch (SocketException ex)
            {
                Console.WriteLine($"SocketException: {ex.Message}");
                Disconnect();
            }
        }

        public bool IsIP6 { get; private set; }

        private Task StartReceivingAsync()
        {
            byte[] buffer = new byte[1024];
            return Task.Factory.StartNew(async () =>
            {
                while (socket.Connected)
                {
                    try
                    {
                        int received = await socket.ReceiveAsync(new ArraySegment<byte>(buffer), SocketFlags.None).ConfigureAwait(false);
                        if (received > 0)
                        {
                            byte[] data = new byte[received];
                            Array.Copy(buffer, data, received);
                            OnReceive?.Invoke(new RemoteInfo
                                { IsTcp = true, Data = data, EndPoint = socket.RemoteEndPoint, IsIpv6 = IsIP6 });
                        }
                        else
                        {
                            break;
                        }
                    }
                    catch (SocketException ex)
                    {
                        Console.WriteLine($"SocketException: {ex.Message}");
                        break;
                    }
                }
                OnDisconnect?.Invoke(new RemoteInfo { IsTcp = true, EndPoint = socket.RemoteEndPoint, IsIpv6 = IsIP6 });
            });
        }

        public async Task SendAsync(byte[] data)
        {
            await socket.SendAsync(new ArraySegment<byte>(data), SocketFlags.None).ConfigureAwait(false);
            OnSend?.Invoke(new RemoteInfo { EndPoint = socket.RemoteEndPoint, Data = data, IsIpv6 = IsIP6 });
        }

        public void Disconnect()
        {
            socket.Close();
        }
    }
}
