﻿using System.Net.Sockets;
using System.Net;
using System.Threading.Tasks;
using System;

namespace EonaCat.Network
{
    public class SocketUdpClient
    {
        public event Action<EndPoint> OnConnect;
        public event Action<byte[]> OnReceive;
        public event Action<EndPoint> OnDisconnect;
        public event Action<EndPoint, byte[]> OnSend;

        private UdpClient udpClient;

        public SocketUdpClient(IPAddress ipAddress, int port)
        {
            CreateUdpClient(ipAddress, port);
        }

        private void CreateUdpClient(IPAddress ipAddress, int port)
        {
            IsIP6 = ipAddress.AddressFamily == AddressFamily.InterNetworkV6;
            udpClient = new UdpClient(ipAddress.AddressFamily);

            if (IsIP6)
            {
                udpClient.Client.SetSocketOption(SocketOptionLevel.IPv6, SocketOptionName.IPv6Only, false);
            }

            udpClient.Client.Bind(new IPEndPoint(IsIP6 ? IPAddress.IPv6Any : IPAddress.Any, 0));

            if (IsMulticastGroupEnabled)
            {
                udpClient.JoinMulticastGroup(ipAddress);
            }

            _ = StartReceivingAsync();
        }

        public bool IsMulticastGroupEnabled { get; set; }

        public bool IsIP6 { get; private set; }

        public SocketUdpClient(string ipAddress, int port)
        {
            if (!IPAddress.TryParse(ipAddress, out IPAddress ip))
            {
                throw new Exception("Invalid ipAddress given");
            }

            CreateUdpClient(ip, port);
        }

        private Task StartReceivingAsync()
        {
            return Task.Factory.StartNew(async () =>
            {
                while (true)
                {
                    try
                    {
                        UdpReceiveResult result = await udpClient.ReceiveAsync();
                        OnReceive?.Invoke(result.Buffer);
                    }
                    catch (SocketException ex)
                    {
                        Console.WriteLine($"SocketException: {ex.Message}");
                        break;
                    }
                }
                OnDisconnect?.Invoke(udpClient.Client.LocalEndPoint);
            });
        }

        public async Task SendTo(EndPoint endPoint, byte[] data)
        {
            if (endPoint is IPEndPoint ipEndPoint)
            {
                await udpClient.SendAsync(data, data.Length, ipEndPoint).ConfigureAwait(false);
                OnSend?.Invoke(endPoint, data);
            }
        }

        public void Disconnect()
        {
            udpClient.Close();
        }
    }
}
