﻿using System.Net.Sockets;
using System.Net;
using System;
using System.Runtime.InteropServices;
using System.Net.NetworkInformation;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    public class SocketUdpServer
    {
        public event Action<RemoteInfo> OnReceive;
        public event Action<RemoteInfo> OnSend;
        public event Action<RemoteInfo> OnDisconnect;
        

        private UdpClient udpClient;

        public bool IsIP6 { get; private set; }

        public SocketUdpServer(IPAddress ipAddress, int port)
        {
            CreateUdpServer(ipAddress, port);
        }

        public SocketUdpServer(string ipAddress, int port)
        {
            if (!IPAddress.TryParse(ipAddress, out IPAddress ip))
            {
                throw new Exception("Invalid ipAddress given");
            }

            CreateUdpServer(ip, port);
        }


        private static void SetConnectionReset(Socket socket)
        {
            if (socket.ProtocolType != ProtocolType.Udp && !RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
            {
                return;
            }

            // Disable ICMP packet shutdown (forcibly closed)
            const int SioUdpConnReset = -1744830452;

            byte[] inValue = { 0, 0, 0, 0 };
            socket.IOControl(SioUdpConnReset, inValue, null);
        }

        private void CreateUdpServer(IPAddress ipAddress, int port)
        {
            IsIP6 = ipAddress.AddressFamily == AddressFamily.InterNetworkV6;
            udpClient = new UdpClient(ipAddress.AddressFamily);
            SetConnectionReset(udpClient.Client);
            
            if (IsIP6)
            {
                udpClient.Client.SetSocketOption(SocketOptionLevel.IPv6, SocketOptionName.IPv6Only, false);
            }

            udpClient.Client.Bind(new IPEndPoint(ipAddress, port));
        }

        public Task StartAsync()
        {
            return Task.Factory.StartNew(async () =>
            {
                while (true)
                {
                    try
                    {
                        UdpReceiveResult result = await udpClient.ReceiveAsync().ConfigureAwait(false);
                        OnReceive?.Invoke(new RemoteInfo()
                        {
                            EndPoint = result.RemoteEndPoint,
                            IsIpv6 = result.RemoteEndPoint.AddressFamily == AddressFamily.InterNetworkV6,
                            Data = result.Buffer
                        });
                    }
                    catch (SocketException ex)
                    {
                        Console.WriteLine($"SocketException: {ex.Message}");
                    }
                }
            });
        }

        public async Task SendToAsync(EndPoint endPoint, byte[] data)
        {
            if (endPoint is IPEndPoint ipEndPoint)
            {
                await udpClient.SendAsync(data, data.Length, ipEndPoint).ConfigureAwait(false);
                OnSend?.Invoke(new RemoteInfo() { EndPoint = endPoint, Data = data, IsIpv6 = endPoint.AddressFamily == AddressFamily.InterNetworkV6 });
            }
        }

        public async Task SendToAllAsync(byte[] data)
        {
            // get all connected clients
            IPGlobalProperties ipProperties = IPGlobalProperties.GetIPGlobalProperties();
            IPEndPoint[] endPoints = ipProperties.GetActiveUdpListeners();
            foreach (IPEndPoint endPoint in endPoints)
            {
                await udpClient.SendAsync(data, data.Length, endPoint).ConfigureAwait(false);
            }
        }
    }
}
